// Code door - Master code + 10 users pin
// Author: Sam Bozman - Garage keypad
// Modified: (c) Szabi 2023
// http://www.dcc-decoders.com

// ************* INSTRUCTIONS **************
// Change Master code: Master code - # - new code - #
// Add new code: Master code - D - new code - D
// Erase all codes: Master code - *
// Open door: Code - A/B
// Cancel operations - C
// ****************************************

#include <Keypad.h>
#include <LiquidCrystal_I2C.h>
#include <EEPROM.h>

LiquidCrystal_I2C lcd(0x3F, 16, 2);  // address of your I2C LCD

long Time = 3000;                // door open time interval [ms]
unsigned long t;                 // stand-by flashing LED
int manual = A0;                 // manual door open pin
int piezoPin = A1;               // pin for piezo transducer
int redLed = 10;                 // stand-by flashing LED pin
int greenLed = 11;               // pin ok LED
int doorPin = 12;                // relay pin

int pinPointer = 0;              // initialize pinPointer variable
const int pinPointerSave = 250;  // dave the pin pointer at EEProm address 250
String pinCode =  "";            // variable to hold Pin Codes (4) digits
String temp = "";                // temporary variable used to compare strings
char pinChar;                    // used to read individual characters from EEPROM
char mf = 'X';                   // used to hold Master Function code ('D' or '#')
char key;                        // used to read in keypad key
boolean result;                  // used to indicate true or false
int frequency = 1047;            // frequency of piezo sound
int k = 4;                       // starting pixel of pinCode
int ledState = LOW;              // ledState used to set the LED
long previousMillis = 0;         // will store last time LED was updated

const byte ROWS = 4;             // four rows
const byte COLS = 4;             // four columns

char hexaKeys[ROWS][COLS] = {
  {'1', '2', '3', 'A'},
  {'4', '5', '6', 'B'},
  {'7', '8', '9', 'C'},
  {'*', '0', '#', 'D'}
};
byte rowPins[ROWS] = {2, 3, 4, 5};  // the row pinouts of the keypad
byte colPins[COLS] = {6, 7, 8, 9};  // the column pinouts of the keypad

Keypad KeyPad = Keypad(makeKeymap(hexaKeys), rowPins, colPins, ROWS, COLS);

void setup() {

  Serial.begin(9600);

  pinMode(doorPin, OUTPUT);
  pinMode(redLed, OUTPUT);
  pinMode(greenLed, OUTPUT);
  pinMode(piezoPin, OUTPUT);
  pinMode(manual, INPUT_PULLUP);

  digitalWrite(doorPin, LOW);       // make sure door switch is off
  digitalWrite(redLed, HIGH);       // turn red led off
  digitalWrite(greenLed, HIGH);     // turn red led off
  KeyPad.setDebounceTime(50);       // default is 50mS
  initializeMaster();               // set Master pinCode

  lcd.init();
  lcd.backlight();                  // backlight ON
  //lcd.noBacklight();              // backlight OFF
  lcd.setCursor(0, 0);
  lcd.print("dcc-decoders.com");
  lcd.setCursor(0, 1);
  lcd.print("Security system");
  delay(3000);
  status(1);
}

void loop() {

  digitalWrite(manual, HIGH);
  if (digitalRead(manual) == LOW) { // manual door open
    openDoor();
  }

  char key = KeyPad.getKey();

  if (key == '0' || key == '1' || key == '2' || key == '3' ||
      key == '4' || key == '5' || key == '6' || key == '7' ||
      key == '8' || key == '9' ) {
    playBeep(1, 50);     // Make short sound
    pinCode[k++] = key;
    lcd.setCursor(k, 1);
    lcd.print("*");     // display '*'
    //lcd.print(key);   // display numbers
  }

  if (key)               // if a key press was detected
  {
    switch (key)         // check to see what was pressed
    {
      case '*':          // erase all Pin Codes
        Serial.println(key);

        switch (mf)      // test value of Master Flag ('D' or '#' or '*')
        {
          case 'X':
            if (checkMaster(pinCode) == true) // is this the Master Pin?
            {
              Serial.println("All user Pin Codes erased!");
              lcd.clear();
              lcd.setCursor(0, 0);
              lcd.print("All codes erased");
              lcd.setCursor(3, 1);
              lcd.print("Users:");
              pinPointer = 0;
              lcd.setCursor(11, 1);
              lcd.print(pinPointer);
              delay(3000);
              EEPROM.write(pinPointerSave, 0); // Erase User Pin Codes
              status(1);
            } else // status(05); // Master Pin Code is NOT Valid!
              break;

          default:      // Was not 'X'
            // if nothing else matches
            status(10); // Unknown Error
            break;
        }
        break;//case '*'

      case 'D': // Add a new User pin code - Key Pressed was 'D'
        Serial.println(key);

        switch (mf) // Test value of Master Flag ('X' or 'D')
        {
          case 'X':
            if (checkMaster(pinCode) == true) // is the Master code valid?
            {
              Serial.println("Master pin code is valid");
              lcd.clear();
              lcd.setCursor(0, 0);
              lcd.print("Master code OK");
              delay(1000);
              lcd.setCursor(0, 0);
              lcd.print("Enter new code");
              lcd.setCursor(0, 1);
              lcd.print("PASS:           ");
              k = 4;
              mf = 'D';
            } else
              status(10); // Unknown Error
            break;

          case 'D': // Master has already been verified
            // Check that new pin code length is correct
            // and if OK then add the new pin Code
            if (pinCode.length() == 4)
              addPinCode(pinCode);
            else if (pinCode.length() != 4)
              status(3);
            break;

          default:
            status(10); // Unknown Error
            break;
        }
        break;

      case '#': // Add a new Master pin code. - Key pressed was '#' (2ND)
        Serial.println(key);

        switch (mf) // Test value of Master Flag ('D' or '#' or '*')
        {
          case 'X':
            if (checkMaster(pinCode) == true) // is the Master code valid?
            {
              Serial.println("Master code is valid");
              lcd.clear();
              lcd.setCursor(2, 0);
              lcd.print("Changing the");
              lcd.setCursor(2, 1);
              lcd.print("master code");
              delay(3000);
              lcd.setCursor(0, 0);
              lcd.print("New Master code");
              lcd.setCursor(0, 1);
              lcd.print("PASS:           ");
              mf = '#'; // Set Master flag
            }
            else status(10); // Master Pin Code is NOT Valid!
            break;

          case '#':// Master has already been verified
            // Check that new pin code length is correct
            // and if OK then add the new pin Code
            if (pinCode.length() == 4) {
              addNewMaster(pinCode);
            } else if (pinCode.length() != 4)
              status(3);
            break;

          default:
            // if nothing else matches
            status(10); // Unknown Error
            break;
        }
        break;

      case 'A':    // If Pin code is valid then toggle door latch
        Serial.println(key);
        if (checkPinCode(pinCode)) openDoor();
        else status(2); // Pin Code invalid!
        break;

      case 'B':    // If Pin code is valid then toggle door latch
        Serial.println(key);
        if (checkPinCode(pinCode)) openDoor();
        else status(2); // Pin Code invalid!
        break;

      case 'C':    // Clear current pinCode entry and start again
        Serial.println(key);
        status(1);
        break;

      default:     // If none of the above then must be a digit but we will double check
        if (key >= 0x30 && key <= 0x39) processDigit(key); // If key is between 0 and 9
        else status(10); // Unknown Error
        break;
    }
  }
  flash_timer();
}

void addNewMaster(String pinIn) {

  if (!checkPinCode(pinIn))
  {
    for (int i = 0; i < 4; i++)
    {
      EEPROM.write(i + 4, pinCode.charAt(i));
    }
    Serial.println("New Master code was successfully added!");
    lcd.clear();
    lcd.setCursor(2, 0);
    lcd.print("Master code");
    lcd.setCursor(2, 1);
    lcd.print("was changed");
    delay(3000);
    status(1);
  }
  else status(8); // New Master matches an existing user code
}

void openDoor() {

  playBeep(2, 100); // Make OK sound
  Serial.println("Pin Code was OK! Door unlocked");
  lcd.clear();
  lcd.setCursor(2, 0);
  lcd.print("Password OK!");
  lcd.setCursor(1, 1);
  lcd.print("Door unlocked");
  digitalWrite(doorPin, HIGH);   // door relay ON
  digitalWrite(greenLed, LOW);   // door LED LED on
  digitalWrite(redLed, HIGH);    // stand-by LED off
  delay(Time);                   // wait x seconds
  digitalWrite(doorPin, LOW);    // door relay OFF
  digitalWrite(greenLed, HIGH);  // door LED LED off
  status(1);
}

void addPinCode(String pinCode) {

  if (!masterReplication(pinCode)) // Check that New Pin Code is not the same as the Master Code
  {
    if (!checkPinCode(pinCode))    // Check that New pin code does not already exist
    {
      pinPointer = EEPROM.read(pinPointerSave); // Read the pinPointer (Max 10)
      if (pinPointer < 10) // Write pinCode to next memory spot
      {
        for (int i = 0; i < 4; i++)
        {
          EEPROM.write(((pinPointer + 2) * 4 + i), pinCode.charAt(i));
        }
        pinPointer++; // Add 1 to pinPointer
        EEPROM.write(pinPointerSave, pinPointer); // Write new pinPonter value back to pinPointerSave address
        Serial.println("New user pin code successfully added");
        lcd.clear();
        lcd.setCursor(1, 0);
        lcd.print("New code added");
        lcd.setCursor(3, 1);
        lcd.print("Users:");
        lcd.setCursor(11, 1);
        lcd.print(pinPointer);
        delay(3000);
        status(1);
      }
      else if (pinPointer >= 10) status(6); // User pin codes at a maximum of 10
    }
    else status(9); // New pin code already exists!
  }
  else status(7); // Pin is teh same with the Master
}

boolean checkPinCode(String pinCodeIn) {

  temp = "";                                   // initialize compare variable
  pinPointer = EEPROM.read(pinPointerSave);    // get the pinPointer (0-9);

  for (int i = 0; i < pinPointer; i++ ) {      // cycle through saved pin code addresses
    for (int j = 0 ; j < 4; j++) {             // cycle through each digit of  current pin code
      pinChar = EEPROM.read((i + 2) * 4 + j);  // read in 4 digits into temp variable
      temp += pinChar;
    }
    if (temp == pinCodeIn) {
      temp = "";
      return true;
    } else {
      temp = "";
    }
  }
  return false;
}

boolean masterReplication(String repCode) {

  temp = "";  // Get  4 char of Master code
  for (int i = 4; i < 8; i++)
  {
    pinChar = (EEPROM.read(i));
    temp += pinChar;
  }
  if (temp == repCode)
  {
    return true;
  }
  return false;
}

void status(int statusIn) {

  switch (statusIn)
  {
    case 1:
      Serial.println("Enter password");
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Enter password");
      lcd.setCursor(0, 1);
      lcd.print("PASS:    ");
      k = 4;
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 100); // Make Error sound
      break;

    case 2:
      Serial.println("Invalid password");
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Invalid password");
      delay(2000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 3:
      Serial.println("Code save failed, try it again!");
      lcd.clear();
      lcd.setCursor(2, 0);
      lcd.print("Status error,");
      lcd.setCursor(2, 1);
      lcd.print("try it again!");
      delay(3000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 4:
      Serial.println("Expected a function key but a number was entered instead!");
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 5:
      Serial.println("Master code is NOT valid!");
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Master code is");
      lcd.setCursor(2, 1);
      lcd.print("not  valid");
      delay(3000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 6:
      Serial.println("Number of codes are at the maximum of 10");
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Number of codes");
      lcd.setCursor(0, 1);
      lcd.print("are at the max");
      delay(3000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 7:
      Serial.println("New User Pin Code matches the current Master Code!");
      lcd.clear();
      lcd.setCursor(2, 0);
      lcd.print("Pin code same");
      lcd.setCursor(0, 1);
      lcd.print("with Master code");
      delay(3000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 8:
      Serial.println("New Master Code matches an existing User Pin Code!");
      lcd.clear();
      lcd.setCursor(0, 0);
      lcd.print("Master code same");
      lcd.setCursor(1, 1);
      lcd.print("with user pin!");
      delay(3000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 9:
      Serial.println("New pin code already exists!");
      lcd.clear();
      lcd.setCursor(2, 0);
      lcd.print("New pin code");
      lcd.setCursor(1, 1);
      lcd.print("already exists");
      delay(3000);
      status(1);
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    case 10:
      Serial.println("Programming Error");
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;

    default:
      Serial.println("Status Error. Reason Unknown!");
      mf = 'X';         // Clear Master Flag
      pinCode = "";     // Clear pinCode
      playBeep(1, 500); // Make Error sound
      break;
  }
}

boolean checkMaster(String mastCode)
{
  if (pinCode.length() == 4)
  {
    temp = "";
    // Get first 4 char of EEProm memory
    for (int i = 4; i < 8; i++)
    {
      pinChar = (EEPROM.read(i));
      temp += pinChar;
    }
    if (temp == mastCode)
    {
      status(1);
      return true;
    }
    else if (temp != mastCode)
    {
      status(5); // Set status to 'Master Pin Code was invalid!'
      return false;
    }
  }
  else if (pinCode.length() != 4)
  {
    status(3);
    return false;
  }
}

void initializeMaster() {

  pinCode = "";
  // Get first 4 char of EEProm memory
  for (int i = 0; i < 4; i++) {
    pinChar = (EEPROM.read(i));
    pinCode += pinChar;
  }
  // If masterCode does not equal 'MAST' then
  // write 'MAST1234' into first 8 bytes of EEProm memory
  if (pinCode != "MAST")
  {
    Serial.println("Master code does NOT exist");
    pinCode = "MAST";
    for (int i = 0; i < 4; i++) {
      EEPROM.write(i, pinCode.charAt(i));
    }
    pinCode = "1234";
    for (int i = 4; i < 8; i++) {
      EEPROM.write(i, pinCode.charAt(i - 4));
    }
    // Since this is a new set we will reset the pinPoniter to '0'
    // and save this value to EEProm address '250'
    pinPointer = 0;
    EEPROM.write(pinPointerSave, pinPointer);
    pinPointer = EEPROM.read(pinPointerSave);
    Serial.println("Initialized pinPointer = ");
    Serial.println(pinPointer);
    pinCode = "";
  }
  else {
    // If this is not a new set then we will load the pinPointer into variable
    Serial.println("Master code DOES exist");
    // Get 4 characters of Master Pin
    pinCode = "";
    for (int i = 4; i < 8; i++) {
      pinChar = (EEPROM.read(i));
      pinCode += pinChar;
    }
    pinPointer = EEPROM.read(pinPointerSave);
    Serial.print("Existing Master pinCode = ");
    Serial.println(pinCode);
    Serial.print("Existing pinPointer = ");
    Serial.println(pinPointer);
    pinCode = "";
  }
  pinCode = "";
}

void processDigit(char keyIn) {

  if (pinCode.length() < 4) {
    pinCode += keyIn; // add digit to pinCode
    if (pinCode.length() == 4)
      Serial.println("Pin Code entered is " + pinCode);
  }
  else status(4);
}

void playBeep(int repeat, int duration) {

  for (int i = 0; i < repeat; i++) {
    tone(piezoPin, frequency, duration);
    delay(duration * 1.30);
  }
}

void flash_timer() {    // stand-by flashing LED

  unsigned long currentMillis = millis();
  if (currentMillis - previousMillis > t) {
    previousMillis = currentMillis;
    if (ledState == HIGH) {
      ledState = LOW;
      t = 100;          // blink interval [ms]
    } else {
      ledState = HIGH;
      t = 3000;         // pause interval [ms]
    }
    digitalWrite(redLed, ledState);
  }
}
